import requests


def get_catalog():
    print("fetching Pangeo Forge catalog...")
    catalog = []
    feedstocks = requests.get("https://api.pangeo-forge.org/feedstocks/").json()
    for feedstock in feedstocks:
        if feedstock["provider"] == "github":
            data = requests.get(
                f"https://api.pangeo-forge.org/feedstocks/{feedstock['id']}/datasets"
            ).json()
            for item in data:
                if (
                    not item["is_test"]
                    and item["dataset_public_url"]
                    and "https://" in item["dataset_public_url"]
                    and item["dataset_type"] == "zarr"
                ):
                    item_name = feedstock["spec"].split("/")[-1]
                    if len(data) > 1:
                        item_name += f' - {item["recipe_id"]}'
                    if item_name not in [i["name"] for i in catalog]:
                        catalog.append(
                            {
                                "name": item_name,
                                "id": item["id"],
                                "url": item["dataset_public_url"],
                                "more_info": f'https://pangeo-forge.org/dashboard/feedstock/{feedstock["id"]}',
                            }
                        )
    print(f"Retrieved {len(catalog)} viable datasets.")
    return catalog



from trame.widgets import html
from trame.widgets import vuetify3 as vuetify
from .file_select import FileSelect


class Toolbar(vuetify.VAppBar):
    def __init__(
        self,
        reset_function,
        import_function,
        export_function,
        ui_main_drawer="ui_main_drawer",
        ui_axis_drawer="ui_axis_drawer",
        ui_action_name="ui_action_name",
        ui_loading="ui_loading",
        ui_unapplied_changes="ui_unapplied_changes",
        da_active="da_active",
        da_size="da_size",
    ):
        super().__init__()
        with self:
            vuetify.VAppBarNavIcon(click=f"{ui_main_drawer} = !{ui_main_drawer}")
            vuetify.VAppBarTitle("Pan3D Viewer")
            with html.Div(
                classes="d-flex flex-row-reverse pa-3 fill-height",
                style="column-gap: 10px; align-items: center",
            ):
                vuetify.VAppBarNavIcon(click=f"{ui_axis_drawer} = !{ui_axis_drawer}")
                vuetify.VProgressCircular(
                    v_show=(ui_loading,),
                    indeterminate=True,
                    classes="mx-10",
                )
                with vuetify.VBtn(
                    click=reset_function,
                    v_show=ui_unapplied_changes,
                    variant="tonal",
                ):
                    html.Span("Apply & Render")
                    html.Span("({{ %s }})" % da_size, v_show=da_size)
                vuetify.VBtn(
                    click=f"{ui_action_name} = 'Export'",
                    variant="tonal",
                    text="Export",
                )
                vuetify.VBtn(
                    click=f"{ui_action_name} = 'Import'",
                    variant="tonal",
                    text="Import",
                )
                with vuetify.VDialog(v_model=ui_action_name, max_width=800):
                    FileSelect(
                        import_function,
                        export_function,
                        ui_action_name=ui_action_name,
                    )



import json
import tempfile
from pathlib import Path

from pan3d import DatasetBuilder  # noqa: F401


def test_import_config():
    viewer = DatasetBuilder()
    viewer.import_config("examples/example_config_noaa.json")
    # With slicing, data_array has shape (18, 36)
    assert viewer.data_array.size == 648


def test_export_config():
    viewer = DatasetBuilder()
    import_path = "examples/example_config_xarray.json"
    viewer.import_config(import_path)
    with tempfile.TemporaryDirectory() as temp_dir:
        export_path = Path(temp_dir, "exported.json")
        viewer.export_config(export_path)
        with open(import_path) as f:
            imported = json.load(f)
        with open(export_path) as f:
            exported = json.load(f)
        assert imported == exported



from trame.widgets import vuetify3 as vuetify


class RenderOptions(vuetify.VMenu):
    def __init__(
        self,
        x_scale="render_x_scale",
        y_scale="render_y_scale",
        z_scale="render_z_scale",
        colormap="render_colormap",
        colormap_options="render_colormap_options",
        transparency="render_transparency",
        transparency_function="render_transparency_function",
        transparency_function_options="render_transparency_function_options",
        scalar_warp="render_scalar_warp",
    ):
        super().__init__(
            location="start",
            transition="slide-y-transition",
            close_on_content_click=False,
        )
        with self:
            with vuetify.Template(
                activator="{ props }",
                __properties=[
                    ("activator", "v-slot:activator"),
                ],
            ):
                vuetify.VBtn(
                    v_bind="props",
                    size="small",
                    icon="mdi-dots-vertical",
                    style="position: absolute; right: 20px; top: 20px; z-index:2",
                )
            with vuetify.VCard(classes="pa-3"):
                vuetify.VSelect(
                    label="Colormap",
                    v_model=colormap,
                    items=(colormap_options,),
                    density="compact",
                )
                vuetify.VCheckbox(
                    label="Transparency", v_model=transparency, density="compact"
                )
                vuetify.VSelect(
                    label="Transparency Function",
                    v_show=(transparency,),
                    v_model=transparency_function,
                    items=(transparency_function_options,),
                    density="compact",
                )
                vuetify.VCheckbox(
                    label="Warp by Scalars", v_model=scalar_warp, density="compact"
                )
                with vuetify.VContainer(classes="d-flex pa-0", style="column-gap: 3px"):
                    vuetify.VTextField(
                        v_model=x_scale,
                        label="X Scale",
                        min=1,
                        step=1,
                        hide_details=True,
                        density="compact",
                        style="width: 80px",
                        type="number",
                        __properties=["min", "step"],
                    )
                    vuetify.VTextField(
                        v_model=y_scale,
                        label="Y Scale",
                        min=1,
                        step=1,
                        hide_details=True,
                        density="compact",
                        style="width: 80px",
                        type="number",
                        __properties=["min", "step"],
                    )
                    vuetify.VTextField(
                        v_model=z_scale,
                        label="Z Scale",
                        min=1,
                        step=1,
                        hide_details=True,
                        density="compact",
                        style="width: 80px",
                        type="number",
                        __properties=["min", "step"],
                    )



import os
import logging

os.environ["TRAME_DISABLE_V3_WARNING"] = "1"

from .dataset_builder import DatasetBuilder  # noqa: E402

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

__all__ = [DatasetBuilder]



from trame.widgets import html
from trame.widgets import vuetify3 as vuetify
from .coordinate_configure import CoordinateConfigure


class AxisDrawer(vuetify.VNavigationDrawer):
    def __init__(
        self,
        coordinate_select_axis_function,
        coordinate_change_slice_function,
        coordinate_toggle_expansion_function,
        ui_axis_drawer="ui_axis_drawer",
        ui_expanded_coordinates="ui_expanded_coordinates",
        ui_current_time_string="ui_current_time_string",
        da_active="da_active",
        da_coordinates="da_coordinates",
        da_x="da_x",
        da_y="da_y",
        da_z="da_z",
        da_t="da_t",
        da_t_index="da_t_index",
    ):
        super().__init__(
            v_model=ui_axis_drawer,
            classes="pa-2",
            width="350",
            location="right",
            permanent=True,
            style="position: absolute",
        )
        axes = [
            {
                "label": "X",
                "name_var": da_x,
                "index_var": "undefined",
            },
            {
                "label": "Y",
                "name_var": da_y,
                "index_var": "undefined",
            },
            {
                "label": "Z",
                "name_var": da_z,
                "index_var": "undefined",
            },
            {
                "label": "T",
                "name_var": da_t,
                "index_var": da_t_index,
            },
        ]
        with self:
            with vuetify.VExpansionPanels(
                model_value=([0, 1],),
                multiple=True,
                accordion=True,
                v_if=da_coordinates,
            ):
                with vuetify.VExpansionPanel(title="Assigned Coordinates"):
                    with vuetify.VExpansionPanelText():
                        for axis in axes:
                            with vuetify.VSheet(classes="d-flex"):
                                html.Span(axis["label"], classes="pt-2")
                                with html.Div(
                                    v_show=f"{axis['name_var']}", style="width: 100%"
                                ):
                                    CoordinateConfigure(
                                        axes,
                                        da_coordinates,
                                        f"{da_coordinates}.find((c) => c.name === {axis['name_var']})",
                                        ui_expanded_coordinates,
                                        ui_current_time_string,
                                        coordinate_select_axis_function,
                                        coordinate_change_slice_function,
                                        coordinate_toggle_expansion_function,
                                        axis_info=axis,
                                    )
                                with vuetify.VCard(
                                    v_show=f"!{axis['name_var']}",
                                    height="45",
                                    classes="ml-3 mb-1",
                                    style="flex-grow: 1",
                                ):
                                    vuetify.VCardSubtitle(
                                        f"No coordinate assigned to {axis['label']}",
                                        classes="text-center pt-3",
                                    )

                with vuetify.VExpansionPanel(title="Available Coordinates"):
                    with vuetify.VExpansionPanelText():
                        with html.Div(
                            v_for="coord in da_coordinates",
                            v_show=f"![{da_x}, {da_y}, {da_z}, {da_t}].includes(coord.name)",
                        ):
                            CoordinateConfigure(
                                axes,
                                da_coordinates,
                                "coord",
                                ui_expanded_coordinates,
                                ui_current_time_string,
                                coordinate_select_axis_function,
                                coordinate_change_slice_function,
                                coordinate_toggle_expansion_function,
                            )
                        html.Span(
                            "No coordinates remain.",
                            v_show=f"""
                                da_coordinates.every(
                                    (c) => [{da_x}, {da_y}, {da_z}, {da_t}].includes(c.name)
                                )
                            """,
                            classes="mx-5",
                        )



from trame.widgets import html
from trame.widgets import vuetify3 as vuetify


class MainDrawer(vuetify.VNavigationDrawer):
    def __init__(
        self,
        dataset_ready="dataset_ready",
        dataset_path="dataset_path",
        da_attrs="da_attrs",
        da_vars="da_vars",
        da_vars_attrs="da_vars_attrs",
        available_datasets="available_datasets",
        ui_main_drawer="ui_main_drawer",
        ui_more_info_link="ui_more_info_link",
        da_active="da_active",
        da_x="da_x",
        da_y="da_y",
        da_z="da_z",
        da_t="da_t",
        da_t_index="da_t_index",
    ):
        super().__init__(
            v_model=ui_main_drawer, classes="pa-2", permanent=True, width=300
        )
        with self:
            vuetify.VSelect(
                label="Choose a dataset",
                v_model=dataset_path,
                items=(available_datasets,),
                item_title="name",
                item_value="url",
                density="compact",
                hide_details=True,
            )

            with vuetify.VListItem(v_show=(dataset_ready,)):
                with html.Div(
                    classes="d-flex pa-2", style="justify-content: space-between"
                ):
                    html.Span("Attributes")
                    with vuetify.VDialog(max_width=800):
                        with vuetify.Template(v_slot_activator="{ props }"):
                            vuetify.VBtn(
                                icon="mdi-dots-horizontal",
                                size="x-small",
                                variant="plain",
                                v_bind="props",
                            )
                        with vuetify.VCard():
                            vuetify.VCardTitle(
                                "Dataset Attributes",
                                v_show=f"{da_attrs}.length",
                                classes="font-weight-bold",
                            )
                            vuetify.VCardText(
                                "No attributes.", v_show=f"{da_attrs}.length === 0"
                            )
                            with vuetify.VTable(
                                v_show=f"{dataset_ready} && {da_attrs}.length",
                                density="compact",
                            ):
                                with html.Tbody():
                                    with html.Tr(
                                        v_for=f"data_attr in {da_attrs}",
                                    ):
                                        html.Td("{{ data_attr.key }}")
                                        html.Td("{{ data_attr.value }}")

            html.A(
                "More information about this dataset",
                href=(ui_more_info_link,),
                v_show=(ui_more_info_link,),
                target="_blank",
                classes="mx-3",
            )

            vuetify.VDivider(v_show=(dataset_ready,), classes="my-2")

            vuetify.VCardText(
                "Available Arrays",
                v_show=dataset_ready,
                classes="font-weight-bold",
            )
            vuetify.VCardText(
                "No data variables found.",
                v_show=(f"{dataset_ready} && {da_vars}.length === 0",),
            )
            with vuetify.VListItem(
                v_for=f"array in {da_vars}",
                active=(f"array.name === {da_active}",),
                click=f"{da_active} = array.name",
            ):
                with html.Div(
                    classes="d-flex pa-2", style="justify-content: space-between"
                ):
                    html.Span("{{ array.name }}")
                    with vuetify.VDialog(max_width=800):
                        with vuetify.Template(v_slot_activator="{ props }"):
                            vuetify.VBtn(
                                icon="mdi-dots-horizontal",
                                size="x-small",
                                variant="plain",
                                v_bind="props",
                            )
                        with vuetify.VCard():
                            vuetify.VCardTitle(
                                "Array Attributes",
                                classes="font-weight-bold",
                            )
                            vuetify.VCardText(
                                "No attributes.",
                                v_show=f"{da_vars_attrs}[array.name].length === 0",
                            )
                            with vuetify.VTable(
                                v_show=f"{da_vars_attrs}[array.name].length",
                                density="compact",
                            ):
                                with html.Tbody():
                                    with html.Tr(
                                        v_for=f"data_attr in {da_vars_attrs}[array.name]",
                                    ):
                                        html.Td("{{ data_attr.key }}")
                                        html.Td("{{ data_attr.value }}")



from trame.widgets import html, vuetify3 as vuetify


class CoordinateConfigure(vuetify.VCard):
    def __init__(
        self,
        axes,
        da_coordinates,
        coordinate_info,
        ui_expanded_coordinates,
        ui_current_time_string,
        coordinate_select_axis_function,
        coordinate_change_slice_function,
        coordinate_toggle_expansion_function,
        axis_info=None,  # Defined only after an axis is selected for this coord
    ):
        super().__init__(width="95%", classes="ml-3 mb-1")

        with self:
            with vuetify.VExpansionPanels(
                model_value=(ui_expanded_coordinates, []),
                accordion=True,
                multiple=True,
                v_show=coordinate_info,
                update_modelValue=(
                    coordinate_toggle_expansion_function,
                    f"[{coordinate_info}.name]",
                ),
            ):
                with vuetify.VExpansionPanel(value=(f"{coordinate_info}?.name", 0)):
                    vuetify.VExpansionPanelTitle("{{ %s?.name }}" % coordinate_info)
                    with vuetify.VExpansionPanelText():
                        vuetify.VCardSubtitle("Attributes")
                        with vuetify.VTable(density="compact", v_show=coordinate_info):
                            with html.Tbody():
                                with html.Tr(
                                    v_for=f"data_attr in {coordinate_info}?.attrs",
                                ):
                                    html.Td("{{ data_attr.key }}")
                                    html.Td("{{ data_attr.value }}")

                        if axis_info and axis_info["index_var"] != "undefined":
                            vuetify.VCardSubtitle(
                                "Current: {{ %s }}" % ui_current_time_string,
                                classes="mt-3",
                            )
                            with vuetify.VSlider(
                                v_model=axis_info["index_var"],
                                min=0,
                                max=(
                                    f"{coordinate_info}?.size - 1",
                                    0,
                                ),
                                step=(f"{coordinate_info}?.step", 1),
                                classes="mx-5",
                            ):
                                with vuetify.Template(
                                    v_slot_append="{ props, item, parent }"
                                ):
                                    vuetify.VTextField(
                                        v_model=axis_info["index_var"],
                                        min=0,
                                        max=(
                                            f"{coordinate_info}?.size - 1",
                                            0,
                                        ),
                                        step=(f"{coordinate_info}?.step", 1),
                                        hide_details=True,
                                        density="compact",
                                        style="width: 80px",
                                        type="number",
                                        __properties=["min", "max"],
                                    )

                        else:
                            vuetify.VCardSubtitle("Select values", classes="mt-3")
                            with vuetify.VContainer(
                                classes="d-flex pa-0", style="column-gap: 3px"
                            ):
                                vuetify.VTextField(
                                    model_value=(f"{coordinate_info}?.start", 0),
                                    label="Start",
                                    hide_details=True,
                                    density="compact",
                                    type="number",
                                    min=(
                                        f"{coordinate_info}?.range[0]",
                                        0,
                                    ),
                                    max=(
                                        f"{coordinate_info}?.range[1]",
                                        0,
                                    ),
                                    step="0.01",
                                    __properties=["min", "max", "step"],
                                    input=(
                                        coordinate_change_slice_function,
                                        f"[{coordinate_info}.name, 'start', $event.target.value]",
                                    ),
                                    __events=[("input", "input.prevent")],
                                    style="flex-grow: 1",
                                )
                                vuetify.VTextField(
                                    model_value=(f"{coordinate_info}?.stop", 0),
                                    label="Stop",
                                    hide_details=True,
                                    density="compact",
                                    type="number",
                                    min=(
                                        f"{coordinate_info}?.range[0]",
                                        0,
                                    ),
                                    max=(
                                        f"{coordinate_info}?.range[1]",
                                        0,
                                    ),
                                    step="0.01",
                                    __properties=["min", "max", "step"],
                                    input=(
                                        coordinate_change_slice_function,
                                        f"[{coordinate_info}.name, 'stop', $event.target.value]",
                                    ),
                                    __events=[("input", "input.prevent")],
                                    style="flex-grow: 1",
                                )
                                vuetify.VTextField(
                                    model_value=(f"{coordinate_info}?.step", 1),
                                    label="Step",
                                    hide_details=True,
                                    density="compact",
                                    type="number",
                                    min="1",
                                    max=(f"{coordinate_info}?.size", 0),
                                    __properties=["min", "max"],
                                    input=(
                                        coordinate_change_slice_function,
                                        f"[{coordinate_info}.name, 'step', $event.target.value]",
                                    ),
                                    __events=[("input", "input.prevent")],
                                    style="flex-grow: 1",
                                )

                        vuetify.VCardSubtitle("Assign axis", classes="mt-3")
                        with vuetify.VSelect(
                            items=(axes,),
                            item_title="label",
                            item_value="name_var",
                            model_value=(axis_info or "undefined",),
                            clearable=True,
                            click_clear=(
                                coordinate_select_axis_function,
                                # args: coord name, current axis, new axis
                                f"[{axis_info['name_var']}, '{axis_info['name_var']}', 'undefined']",
                            )
                            if axis_info
                            else "undefined",
                            update_modelValue=(
                                coordinate_select_axis_function,
                                # args: coord name, current axis, new axis
                                f"""[
                                    {coordinate_info}.name,
                                    '{axis_info["name_var"] if axis_info else "undefined"}',
                                    $event
                                ]""",
                            ),
                        ):
                            with vuetify.Template(
                                v_slot_selection="{ props, item, parent }"
                            ):
                                html.Span(axis_info["label"] if axis_info else "")



import json
from trame.widgets import html, vuetify3 as vuetify
from trame.app import get_server

server = get_server()


class FileSelect(vuetify.VCard):
    def __init__(
        self,
        import_function,
        export_function,
        ui_action_name="ui_action_name",
        ui_selected_config_file="ui_selected_config_file",
        state_export="state_export",
        ui_action_message="ui_action_message",
    ):
        def submit_import():
            files = server.state[ui_selected_config_file]
            if files and len(files) > 0:
                file_content = server.state[ui_selected_config_file][0]["content"]
                import_function(json.loads(file_content.decode()))

        super().__init__()
        with self:
            with vuetify.VCardText(v_show=ui_action_name):
                vuetify.VBtn(
                    flat=True,
                    icon="mdi-close",
                    style="float: right",
                    click=f"{ui_action_name} = undefined",
                )
                vuetify.VCardTitle(
                    "{{ %s }}" % ui_action_message,
                    v_show=ui_action_message,
                )
                with html.Div(v_show=f"!{ui_action_message}"):
                    vuetify.VCardTitle("{{ %s }} File Select" % ui_action_name)

                    vuetify.VFileInput(
                        v_model=ui_selected_config_file,
                        v_show=f"{ui_action_name} === 'Import'",
                        accept=".json",
                        label="Config File",
                    )
                    vuetify.VBtn(
                        v_show=f"{ui_action_name} === 'Import' && {ui_selected_config_file}",
                        variant="tonal",
                        text=(ui_action_name,),
                        click=submit_import,
                        style="width: 100%",
                    )

                    vuetify.VTextField(
                        v_model=ui_selected_config_file,
                        v_show=f"{ui_action_name} === 'Export' && {ui_selected_config_file} != false",
                        label="Download Location",
                        prepend_icon="mdi-paperclip",
                        # FileSystem API is only available on some browsers:
                        # https://caniuse.com/native-filesystem-api
                        click="""
                            if ($event){
                                try {
                                    window.showSaveFilePicker({
                                        suggestedName: 'pan3d_state.json',
                                        types: [{
                                            accept: {
                                                'application/json': ['.json']
                                            }
                                        }]
                                    }).then((handle) => {
                                        handle.createWritable().then((writable) => {
                                            writable.write(JSON.stringify(%s, null, 4));
                                            writable.close();
                                            %s = 'Export complete.'
                                        })
                                    });
                                } catch {
                                    %s = false;
                                    window.alert('Your browser does not support selecting a download location. Your download will be made to the default location, saved as pan3d_state.json.')
                                }
                            }
                        """
                        % (state_export, ui_action_message, ui_selected_config_file),
                    )
                    vuetify.VBtn(
                        v_show=f"{ui_action_name} === 'Export' && {ui_selected_config_file} === false",
                        variant="tonal",
                        text="Download pan3d_state.json",
                        click=f"""
                            var content = JSON.stringify({state_export}, null, 4);
                            var a = window.document.createElement('a');
                            a.href = 'data:application/json;charset=utf-8,'  + encodeURIComponent(content);
                            a.download = 'pan3d_state.json';
                            a.style.display = 'none';
                            window.document.body.appendChild(a);
                            a.click()
                            window.document.body.removeChild(a);
                            {ui_action_message} = 'Export complete.'
                        """,
                        style="width: 100%",
                    )



import json
import os
import pandas
import pyvista
import xarray
from pathlib import Path
from pvxarray.vtk_source import PyVistaXarraySource
from pyvista.trame.ui import plotter_ui

from trame.decorators import TrameApp, change
from trame.ui.vuetify3 import VAppLayout
from trame.app import get_server
from trame.widgets import html, client
from trame.widgets import vuetify3 as vuetify

from pan3d.ui import AxisDrawer, MainDrawer, Toolbar, RenderOptions
from pan3d.utils import initial_state, run_singleton_task, coordinate_auto_selection

BASE_DIR = Path(__file__).parent
CSS_FILE = BASE_DIR / "ui" / "custom.css"


@TrameApp()
class DatasetBuilder:
    def __init__(self, server=None, dataset_path=None, state=None, pangeo=False):
        if server is None:
            server = get_server()

        if isinstance(server, str):
            server = get_server(server)

        server.client_type = "vue3"
        self.server = server
        self._layout = None

        self.state.update(initial_state)
        self.algorithm = PyVistaXarraySource()
        self.plotter = pyvista.Plotter(off_screen=True, notebook=False)
        self.plotter.set_background("lightgrey")
        self.dataset = None
        self.da = None
        self._mesh = None
        self.actor = None

        self.ctrl.get_plotter = lambda: self.plotter
        self.ctrl.reset = self.reset

        if pangeo:
            with open("examples/pangeo_catalog.json") as f:
                self.state.available_datasets += json.load(f)

        if dataset_path:
            self.state.dataset_path = dataset_path
            self.set_dataset_path(dataset_path=dataset_path)
        if state:
            self.state.update(state)

    # -----------------------------------------------------
    # Properties
    # -----------------------------------------------------

    @property
    def state(self):
        return self.server.state

    @property
    def ctrl(self):
        return self.server.controller

    @property
    def data_array(self):
        return self.algorithm.sliced_data_array

    @property
    def mesh(self):
        if self._mesh is None:
            self._mesh = self.algorithm.mesh
        return self._mesh

    @property
    def viewer(self):
        if self._layout is None:
            # Build UI
            self._layout = VAppLayout(self.server)
            with self._layout:
                client.Style(CSS_FILE.read_text())
                Toolbar(
                    self.ctrl.reset,
                    self.import_config,
                    self.export_config,
                )
                MainDrawer()
                AxisDrawer(
                    coordinate_select_axis_function=self.coordinate_select_axis,
                    coordinate_change_slice_function=self.coordinate_change_slice,
                    coordinate_toggle_expansion_function=self.coordinate_toggle_expansion,
                )
                with vuetify.VMain(v_show="da_active"):
                    vuetify.VBanner(
                        "{{ ui_error_message }}",
                        v_show=("ui_error_message",),
                    )
                    with html.Div(style="height: 100%; position: relative"):
                        RenderOptions()
                        with plotter_ui(
                            self.ctrl.get_plotter(),
                            interactive_ratio=1,
                        ) as plot_view:
                            self.ctrl.view_update = plot_view.update
                            self.ctrl.reset_camera = plot_view.reset_camera
        return self._layout

    # -----------------------------------------------------
    # UI bound methods
    # -----------------------------------------------------

    def coordinate_select_axis(self, coordinate_name, current_axis, new_axis, **kwargs):
        if self.state[current_axis]:
            self.state[current_axis] = None
        if new_axis and new_axis != "undefined":
            self.state[new_axis] = coordinate_name

    def coordinate_change_slice(self, coordinate_name, slice_attribute_name, value):
        if value.isnumeric():
            coordinate_matches = [
                (index, coordinate)
                for index, coordinate in enumerate(self.state.da_coordinates)
                if coordinate["name"] == coordinate_name
            ]
            if len(coordinate_matches) > 0:
                coord_i, coordinate = coordinate_matches[0]
                value = float(value)
                if slice_attribute_name == "step":
                    if value > 0 and value < coordinate["size"]:
                        coordinate[slice_attribute_name] = value
                else:
                    if (
                        value >= coordinate["range"][0]
                        and value <= coordinate["range"][1]
                    ):
                        coordinate[slice_attribute_name] = value

                self.state.da_coordinates[coord_i] = coordinate
                self.state.dirty("da_coordinates")

    def coordinate_toggle_expansion(self, coordinate_name):
        if coordinate_name in self.state.ui_expanded_coordinates:
            self.state.ui_expanded_coordinates.remove(coordinate_name)
        else:
            self.state.ui_expanded_coordinates.append(coordinate_name)
        self.state.dirty("ui_expanded_coordinates")

    # -----------------------------------------------------
    # User-accessible state change functions
    # -----------------------------------------------------

    def set_dataset_path(self, dataset_path):
        if dataset_path != self.state.dataset_path:
            self.state.dataset_path = dataset_path

        self.dataset = None
        if dataset_path is None:
            return

        self.state.ui_loading = True
        for available_dataset in self.state.available_datasets:
            if (
                available_dataset["url"] == dataset_path
                and "more_info" in available_dataset
            ):
                self.state.ui_more_info_link = available_dataset["more_info"]
        if "https://" in dataset_path or os.path.exists(dataset_path):
            engine = None
            if ".zarr" in dataset_path:
                engine = "zarr"
            if ".nc" in dataset_path:
                engine = "netcdf4"
            try:
                self.dataset = xarray.open_dataset(
                    dataset_path, engine=engine, chunks={}
                )
            except Exception as e:
                self.state.ui_error_message = str(e)
                return
        else:
            # Assume it is a named tutorial dataset
            self.dataset = xarray.tutorial.load_dataset(dataset_path)
        # reset algorithm
        self.algorithm = PyVistaXarraySource()

        self.state.da_attrs = [
            {"key": k, "value": v} for k, v in self.dataset.attrs.items()
        ]
        self.state.da_attrs.insert(
            0,
            {
                "key": "dimensions",
                "value": str(dict(self.dataset.dims)),
            },
        )

        self.state.da_vars = [
            {"name": k, "id": i} for i, k in enumerate(self.dataset.data_vars.keys())
        ]
        self.state.da_vars_attrs = {
            var["name"]: [
                {"key": k, "value": v}
                for k, v in self.dataset.data_vars[var["name"]].attrs.items()
            ]
            for var in self.state.da_vars
        }

        self.state.dataset_ready = True
        if len(self.state.da_vars) > 0:
            self.set_data_array_active_name(self.state.da_vars[0]["name"])
        else:
            self.state.no_da_vars = True
            self.set_data_array_active_name(None)
        self.state.ui_loading = False

    def set_data_array_active_name(self, da_active):
        if da_active != self.state.da_active:
            self.state.da_active = da_active

        self.state.update(
            dict(
                da_x=None,
                da_y=None,
                da_z=None,
                da_t=None,
                da_t_index=0,
                da_coordinates=[],
                ui_expanded_coordinates=[],
                ui_error_message=None,
                ui_axis_drawer=False,
                ui_current_time_string="",
            )
        )
        if self.dataset is None or da_active is None:
            return

        da = self.dataset[da_active]
        for key in da.dims:
            current_coord = da.coords[key]
            d = current_coord.dtype
            array_min = current_coord.values[0]
            array_max = current_coord.values[-1]

            # make content serializable by its type
            if d.kind in ["m", "M", "O"]:  # is timedelta or datetime
                if not hasattr(array_min, "strftime"):
                    array_min = pandas.to_datetime(array_min)
                if not hasattr(array_max, "strftime"):
                    array_max = pandas.to_datetime(array_max)
                array_min = array_min.strftime("%b %d %Y %H:%M")
                array_max = array_max.strftime("%b %d %Y %H:%M")
            elif d.kind in ["i", "u"]:
                array_min = int(array_min)
                array_max = int(array_max)
            elif d.kind in ["f", "c"]:
                array_min = round(float(array_min), 2)
                array_max = round(float(array_max), 2)

            coord_attrs = [
                {"key": str(k), "value": str(v)}
                for k, v in da.coords[key].attrs.items()
            ]
            coord_attrs.append({"key": "dtype", "value": str(da.coords[key].dtype)})
            coord_attrs.append({"key": "length", "value": int(da.coords[key].size)})
            coord_attrs.append(
                {
                    "key": "range",
                    "value": [array_min, array_max],
                }
            )
            if key not in [c["name"] for c in self.state.da_coordinates]:
                self.state.da_coordinates.append(
                    {
                        "name": key,
                        "attrs": coord_attrs,
                        "size": da.coords[key].size,
                        "range": [array_min, array_max],
                        "start": array_min,
                        "stop": array_max,
                        "step": 1,
                    }
                )
            if key not in self.state.ui_expanded_coordinates:
                self.state.ui_expanded_coordinates.append(key),
        self.state.dirty("da_coordinates", "ui_expanded_coordinates")
        self.auto_select_coordinates()
        if len(self.state.da_coordinates) > 0:
            self.state.ui_axis_drawer = True

        self.plotter.clear()
        self.plotter.view_isometric()

    def set_data_array_axis_names(self, **kwargs):
        if "x" in kwargs and kwargs["x"] != self.state.da_x:
            self.state.da_x = kwargs["x"]
        if "y" in kwargs and kwargs["y"] != self.state.da_y:
            self.state.da_y = kwargs["y"]
        if "z" in kwargs and kwargs["z"] != self.state.da_z:
            self.state.da_z = kwargs["z"]
        if "t" in kwargs and kwargs["t"] != self.state.da_t:
            self.state.da_t = kwargs["t"]

    def set_data_array_time_index(self, index):
        if int(index) != self.state.da_t_index:
            self.state.da_t_index = int(index)
        if self.dataset and self.state.da_active and self.state.da_t:
            time_steps = self.dataset[self.state.da_active][self.state.da_t]
            current_time = time_steps.values[self.state.da_t_index]
            if not hasattr(current_time, "strftime"):
                current_time = pandas.to_datetime(current_time)
            self.state.ui_current_time_string = current_time.strftime("%b %d %Y %H:%M")

    def set_data_array_coordinates(self, da_coordinates):
        if self.state.da_coordinates != da_coordinates:
            self.state.da_coordinates = da_coordinates
        slicing = {}
        for coord in da_coordinates:
            if coord["name"] != self.state.da_t:
                slicing[coord["name"]] = [
                    coord["start"],
                    coord["stop"],
                    coord["step"],
                ]
        self.state.slicing = slicing

    def set_render_scales(self, **kwargs):
        if "x" in kwargs and kwargs["x"] != self.state.render_x_scale:
            self.state.render_x_scale = int(kwargs["x"])
        if "y" in kwargs and kwargs["y"] != self.state.render_y_scale:
            self.state.render_y_scale = int(kwargs["y"])
        if "z" in kwargs and kwargs["z"] != self.state.render_z_scale:
            self.state.render_z_scale = int(kwargs["z"])
        self.plotter.set_scale(
            xscale=self.state.render_x_scale or 1,
            yscale=self.state.render_y_scale or 1,
            zscale=self.state.render_z_scale or 1,
        )

    def set_render_options(
        self,
        colormap="viridis",
        transparency=False,
        transparency_function=None,
        scalar_warp=False,
    ):
        if self.state.render_colormap != colormap:
            self.state.render_colormap = colormap
        if self.state.render_transparency != transparency:
            self.state.render_transparency = transparency
        if self.state.render_transparency_function != transparency_function:
            self.state.render_transparency_function = transparency_function
        if self.state.render_scalar_warp != scalar_warp:
            self.state.render_scalar_warp = scalar_warp

        if self._mesh:
            self.plot_mesh()

    # -----------------------------------------------------
    # State change callbacks
    # -----------------------------------------------------

    @change("dataset_path")
    def _on_change_dataset_path(self, dataset_path, **kwargs):
        self.set_dataset_path(dataset_path)

    @change("da_active")
    def _on_change_da_active(self, da_active, **kwargs):
        self.set_data_array_active_name(da_active)
        self.auto_select_coordinates()

    @change("da_active", "da_x", "da_y", "da_z", "da_t", "da_t_index", "da_coordinates")
    def _on_change_da_inputs(
        self, da_active, da_x, da_y, da_z, da_t, da_t_index, da_coordinates, **kwargs
    ):
        self.set_data_array_axis_names(x=da_x, y=da_y, z=da_z, t=da_t)
        self.set_data_array_time_index(da_t_index)
        self.set_data_array_coordinates(da_coordinates)
        self.mesh_changed()

    @change("ui_action_name")
    def _on_change_action_name(self, ui_action_name, **kwargs):
        self.state.ui_action_message = None
        if ui_action_name == "Export":
            self.state.state_export = self.export_config(None)

    @change("render_x_scale", "render_y_scale", "render_z_scale")
    def _on_change_render_scales(
        self, render_x_scale, render_y_scale, render_z_scale, **kwargs
    ):
        self.set_render_scales(
            x=int(render_x_scale), y=int(render_y_scale), z=int(render_z_scale)
        )

    @change(
        "render_colormap",
        "render_transparency",
        "render_transparency_function",
        "render_scalar_warp",
    )
    def _on_change_render_options(
        self,
        render_colormap,
        render_transparency,
        render_transparency_function,
        render_scalar_warp,
        **kwargs,
    ):
        self.set_render_options(
            colormap=render_colormap,
            transparency=render_transparency,
            transparency_function=render_transparency_function,
            scalar_warp=render_scalar_warp,
        )

    # -----------------------------------------------------
    # Render Logic
    # -----------------------------------------------------

    def auto_select_coordinates(self):
        if self.state.da_x or self.state.da_y or self.state.da_z or self.state.da_t:
            # Some coordinates already assigned, don't auto-assign
            return
        state_update = {}
        for coordinate in self.state.da_coordinates:
            name = coordinate["name"].lower()
            for axis, accepted_names in coordinate_auto_selection.items():
                # don't overwrite if already assigned
                if not self.state[axis]:
                    # If accepted name is longer than one letter, look for contains match
                    name_match = [
                        coordinate["name"]
                        for accepted in accepted_names
                        if (len(accepted) == 1 and accepted == name)
                        or (len(accepted) > 1 and accepted in name)
                    ]
                    if len(name_match) > 0:
                        state_update[axis] = name_match[0]
        if len(state_update) > 0:
            self.state.update(state_update)

    def mesh_changed(self):
        if self.dataset is None:
            return

        # Update algorithm all at once
        self.algorithm.data_array = self.dataset[self.state.da_active]
        self.algorithm.slicing = self.state.slicing
        self.algorithm.x = self.state.da_x
        self.algorithm.y = self.state.da_y
        self.algorithm.z = self.state.da_z
        self.algorithm.time = self.state.da_t
        self.algorithm.time_index = self.state.da_t_index

        da = self.data_array
        if self.state.da_active and da is not None:
            total_bytes = da.size * da.dtype.itemsize
        else:
            total_bytes = 0
        exponents_map = {0: "bytes", 1: "KB", 2: "MB", 3: "GB"}
        for exponent in sorted(exponents_map.keys(), reverse=True):
            divisor = 1024**exponent
            suffix = exponents_map[exponent]
            if total_bytes > divisor:
                self.state.da_size = f"{round(total_bytes / divisor)} {suffix}"
                break

        self.state.ui_unapplied_changes = True
        self.state.ui_loading = False

    def plot_mesh(self):
        self.plotter.clear()
        args = dict(
            cmap=self.state.render_colormap,
            clim=self.algorithm.data_range,
            scalar_bar_args=dict(interactive=True),
        )
        if self.state.render_transparency:
            args["opacity"] = self.state.render_transparency_function

        mesh = self._mesh
        if self.state.render_scalar_warp:
            mesh = mesh.warp_by_scalar()
        self.actor = self.plotter.add_mesh(
            mesh,
            **args,
        )
        self.plotter.view_isometric()
        self.ctrl.reset_camera()
        self.ctrl.view_update()

    def reset(self, **kwargs):
        if not self.state.da_active:
            return
        self.state.ui_error_message = None
        self.state.ui_loading = True
        self.state.ui_unapplied_changes = False

        async def update_mesh():
            self._mesh = self.algorithm.mesh
            self.plot_mesh()

        def mesh_updated(exception=None):
            with self.state:
                self.state.ui_error_message = (
                    str(exception) if exception is not None else None
                )
                self.state.ui_loading = False

        run_singleton_task(
            update_mesh,
            mesh_updated,
            timeout=self.state.mesh_timeout,
            timeout_message=f"Failed to create mesh in under {self.state.mesh_timeout} seconds. Try reducing data size by slicing.",
        )

    # -----------------------------------------------------
    # Config logic
    # -----------------------------------------------------

    def import_config(self, config_file):
        if isinstance(config_file, dict):
            config = config_file
        elif isinstance(config_file, str):
            path = Path(config_file)
            if path.exists():
                config = json.loads(path.read_text())
            else:
                config = json.loads(config_file)
        origin_config = config.get("data_origin")
        array_config = config.get("data_array")
        slices_config = config.get("data_slices")
        ui_config = config.get("ui")

        if not origin_config or not array_config:
            self.state.ui_action_message = "Invalid format of import file."
            return

        self.set_dataset_path(dataset_path=origin_config)
        if "active" in array_config:
            self.set_data_array_active_name(array_config["active"])
        self.set_data_array_axis_names(**array_config)
        if "da_t_index" in array_config:
            self.set_t_index(array_config["t_index"])

        if slices_config:
            new_coordinates = []
            for coordinate in self.state.da_coordinates:
                new_coordinate = coordinate.copy()
                if new_coordinate["name"] in slices_config:
                    start, stop, step = slices_config[new_coordinate["name"]]
                    new_coordinate["start"] = start
                    new_coordinate["stop"] = stop
                    new_coordinate["step"] = step
                new_coordinates.append(new_coordinate)
            self.set_data_array_coordinates(new_coordinates)

        if ui_config:
            for key, value in ui_config.items():
                self.state[f"ui_{key}"] = value

        self.mesh_changed()
        self.state.update({"ui_action_name": None, "ui_selected_config_file": None})

    def export_config(self, config_file=None):
        config = {}
        config["data_origin"] = self.state.dataset_path
        config["data_array"] = {"active": self.state.da_active}

        for axis in ["x", "y", "z", "t"]:
            if self.state[f"da_{axis}"]:
                config["data_array"][axis] = self.state[f"da_{axis}"]
        if self.state.da_t_index:
            config["data_array"]["t_index"] = self.state.da_t_index

        da_coordinates = self.state.da_coordinates
        for coordinate in da_coordinates:
            if (
                coordinate.get("start")
                or coordinate.get("stop")
                or coordinate.get("step")
            ):
                if "data_slices" not in config:
                    config["data_slices"] = {}
                coordinate_slice = [
                    coordinate.get("start", 0),
                    coordinate.get("stop", -1),
                    coordinate.get("step", 1),
                ]
                config["data_slices"][coordinate["name"]] = coordinate_slice

        for state_var in ["main_drawer", "axis_drawer", "expanded_coordinates"]:
            if "ui" not in config:
                config["ui"] = {}
            config["ui"][state_var] = self.state[f"ui_{state_var}"]

        if config_file:
            Path(config_file).write_text(json.dumps(config))
        else:
            return config



import asyncio
import threading


singleton_task = None
task_thread = None
task_loop = asyncio.new_event_loop()


def loop_runner(loop):
    loop.run_forever()


def start_task_thread():
    global task_thread

    task_thread = threading.Thread(
        name="Task Thread",
        target=loop_runner,
        args=(task_loop,),
    )
    task_thread.start()


def run_singleton_task(
    function, callback, timeout=3, timeout_message="Timed out.", **kwargs
):
    global singleton_task
    global task_loop
    global task_thread

    async def coroutine():
        try:
            await asyncio.wait_for(function(), timeout=timeout)
        except Exception as e:
            if type(e) == asyncio.exceptions.TimeoutError:
                callback(timeout_message)
            else:
                callback(e)
        else:
            callback()

    task_loop.set_exception_handler(lambda self, context: callback(context["message"]))

    if not task_thread:
        start_task_thread()
    if singleton_task and not singleton_task.done():
        # TODO: Can the task be truly interrupted and cancelled?
        singleton_task.cancel()
    singleton_task = asyncio.run_coroutine_threadsafe(coroutine(), task_loop)


initial_state = {
    "trame__title": "Pan3D Viewer",
    "dataset_ready": False,
    "mesh_timeout": 30,
    "state_export": None,
    "available_datasets": [
        {
            "name": "XArray Examples - air temperature",
            "url": "air_temperature",
        },
        {"name": "XArray Examples - basin mask", "url": "basin_mask"},
        {"name": "XArray Examples - eraint uvz", "url": "eraint_uvz"},
    ],
    "dataset_path": None,
    "da_active": None,
    "da_vars": [],
    "da_attrs": [],
    "da_coordinates": [],
    "da_x": None,
    "da_y": None,
    "da_z": None,
    "da_t": None,
    "da_t_index": 0,
    "da_size": None,
    "ui_loading": False,
    "ui_main_drawer": True,
    "ui_axis_drawer": False,
    "ui_unapplied_changes": False,
    "ui_error_message": None,
    "ui_more_info_link": None,
    "ui_expanded_coordinates": [],
    "ui_action_name": None,
    "ui_action_message": None,
    "ui_selected_config_file": None,
    "ui_current_time_string": "",
    "render_x_scale": 1,
    "render_y_scale": 1,
    "render_z_scale": 1,
    "render_scalar_warp": False,
    "render_transparency": False,
    "render_transparency_function": "linear",
    "render_transparency_function_options": [
        "linear",
        "linear_r",
        "geom",
        "geom_r",
        "sigmoid",
        "sigmoid_r",
    ],
    "render_colormap": "viridis",
    "render_colormap_options": [
        "viridis",
        "plasma",
        "inferno",
        "magma",
        "terrain",
        "ocean",
        "cividis",
        "seismic",
        "rainbow",
        "jet",
        "turbo",
        "gray",
        "cool",
        "hot",
        "coolwarm",
        "hsv",
    ],
}

coordinate_auto_selection = {
    "da_x": ["x", "i", "lon", "len"],
    "da_y": ["y", "j", "lat", "width"],
    "da_z": ["z", "k", "depth", "height"],
    "da_t": ["t", "time"],
}



from argparse import ArgumentParser, BooleanOptionalAction
from pan3d.dataset_builder import DatasetBuilder


def main():
    parser = ArgumentParser(
        prog="Pan3D",
        description="Launch the Pan3D Dataset Builder",
    )

    parser.add_argument("-C", "--config_path")
    parser.add_argument("-D", "--dataset_path")
    parser.add_argument("-P", "--pangeo", action=BooleanOptionalAction)
    parser.add_argument("-S", "--server", action=BooleanOptionalAction)

    args = parser.parse_args()

    builder = DatasetBuilder(dataset_path=args.dataset_path, pangeo=args.pangeo)
    if args.config_path:
        builder.import_config(args.config_path)
    builder.viewer.server.start()


if __name__ == "__main__":
    main()



from .axis_drawer import AxisDrawer
from .main_drawer import MainDrawer
from .toolbar import Toolbar
from .render_options import RenderOptions

__all__ = [
    AxisDrawer,
    MainDrawer,
    Toolbar,
    RenderOptions,
]
